#ifndef TRANSLIB_H
#define TRANSLIB_H

#include <vector>
#include <string>

struct Vector4
{
  union {
    struct {
      float x, y, z, w;
    };
    float v[4];
  };
  Vector4(float x_, float y_, float z_, float w_):
    x(x_), y(y_), z(z_), w(w_) {};
  Vector4():
    Vector4(0, 0, 0, 0) {};
  float& operator[](int index);
  const float& operator[](int index) const;
  bool operator==(const Vector4& other) const;
  bool operator!=(const Vector4& other) const;
  Vector4& operator=(const Vector4& other);
  Vector4 operator+(const Vector4& other) const;
  Vector4 operator-(const Vector4& other) const;
  Vector4 operator*(float factor) const;
  Vector4 operator/(float factor) const;
  float operator&(const Vector4& other) const;
  Vector4 operator-() const;
  Vector4& operator+=(const Vector4& other);
  Vector4& operator-=(const Vector4& other);
  Vector4& operator*=(float factor);
  Vector4& operator/=(float factor);
};

struct Matrix4
{
  Vector4 a1, a2, a3, a4;
  Matrix4(Vector4 a1_, Vector4 a2_, Vector4 a3_, Vector4 a4_):
    a1(a1_), a2(a2_), a3(a3_), a4(a4_) {};
  Matrix4(float a11, float a12, float a13, float a14,
         float a21, float a22, float a23, float a24,
         float a31, float a32, float a33, float a34,
         float a41, float a42, float a43, float a44):
    Matrix4(Vector4(a11, a12, a13, a14),
           Vector4(a21, a22, a23, a24),
           Vector4(a31, a32, a33, a34),
           Vector4(a41, a42, a43, a44)) {};
  Matrix4():
    Matrix4(Vector4(), Vector4(), Vector4(), Vector4()) {};
  Matrix4(float factor):
    Matrix4(
      factor , 0      , 0      , 0        ,
      0      , factor , 0      , 0        ,
      0      , 0      , factor , 0        ,
      0      , 0      , 0      , factor) {};
  Matrix4(const float *values);

  Vector4& operator[](int index);
  void toGl() const;
  const Vector4& operator[](int index) const;
  bool operator==(const Matrix4& other) const;
  bool operator!=(const Matrix4& other) const;
  Matrix4& operator=(const Matrix4& other);
  Matrix4 operator*(const Matrix4& other) const;
  Vector4 operator*(const Vector4& v) const;
  Matrix4 operator*(float factor) const;
  Matrix4 operator+(const Matrix4& other) const;
  Matrix4 operator-(const Matrix4& other) const;
  Matrix4& operator*=(const Matrix4& other);
  Matrix4& operator*=(float factor);
  Matrix4& operator+=(const Matrix4& other);
  Matrix4& operator-=(const Matrix4& other);
  Matrix4 transpose() const;
  float determinant() const;
  Matrix4 inverse() const;
  Matrix4 to3() const;
};

struct Vector: public Vector4
{
  Vector(float x_, float y_, float z_):
    Vector4(x_, y_, z_, 1) {};
  Vector():
    Vector(0, 0, 0) {};
  Vector(const Vector4& vec4):
    Vector(vec4.x, vec4.y, vec4.z) {};
  Vector operator*(const Vector& other) const;
  Vector operator*(float factor) const;
  Vector operator/(float factor) const;
  Vector& operator*=(const Vector& other);
  Vector& operator*=(float factor);
  Vector& operator/=(float factor);
  float operator&(const Vector& other) const;
  float length() const;
  Vector normal() const;
  Vector& normalize();
  float distance(const Vector& other) const;
  static Vector average(const std::vector<Vector>& vectors);
};

class Transformable
{
  void updateNormalTransformations();
  protected:
    Matrix4 transformations;
    Matrix4 normalTransformations;
    Matrix4 translationMatrix(const Vector& offset) const;
    Matrix4 rotationMatrix(const Vector& axis, float angle, const Vector& center) const;
    Matrix4 scalingMatrix(const Vector& factors) const;
  public:
    Transformable();
    virtual void applyTransformations() =0;
    virtual void move(const Vector& offset);
    virtual void scale(const Vector& factors);
    virtual void rotate(const Vector& axis, float angle);
    virtual void rotate(const Vector& axis, float angle, const Vector& center);
    void clearTransformations();
};
#endif
