#ifndef UTIL_H
#define UTIL_H

#include <GL/glut.h>
#include <vector>
#include <set>
#include <memory>
#include "translib.h"

#define SMOOTH_SHADING 0
#define FLAT_SHADING 1

class Paintable
{
  protected:
    virtual void beforePaint(int shadingType);
    virtual void paintMethod(int shadingType) =0;
    virtual void afterPaint(int shadingType);
  public:
    void paint(int shadingType);
};

class Vertex : public Paintable
{
  protected:
    void paintMethod(int shadingType);
  public:
    Vector point;
    Vector normal;
    Vertex(GLfloat x, GLfloat y, GLfloat z):
      point(x, y, z), normal(0, 0, 0) {};
    Vertex(const Vertex& other):
      point(other.point.x, other.point.y, other.point.z),
      normal(other.normal.x, other.normal.y, other.normal.z) {};
};

class Face : public Paintable
{
  std::vector<std::shared_ptr<Vertex> > vertices;
  Vector normal;
  protected:
    void paintMethod(int shadingType);
  public:
    bool useTexture;
    GLuint texture;
    Vector getNormal() const;
    const Vector& getBase() const;
    void setNormal();
    float distanceTo(const Vector& point) const;
    Vector closestPointTo(const Vector& point) const;
    Face clone() const;
    void sumNormals();
    Face& addVertex(std::shared_ptr<Vertex> vertex);
    std::vector<std::shared_ptr<Vertex> >& getVertices();
};

class FaceGroup : public Paintable, public std::enable_shared_from_this<FaceGroup>
{
  protected:
    std::vector<Face> faces;
    void paintMethod(int shadingType);
    void afterPaint(int shadingType);
    virtual std::shared_ptr<FaceGroup> getFaceGroupPointer() =0;
  public:
    void cloneInto(std::shared_ptr<FaceGroup> faceGroup) const;
    virtual std::shared_ptr<FaceGroup> clone() const =0;
    std::shared_ptr<FaceGroup> addFace(Face face);
    std::vector<Face>& getFaces();
    const std::vector<Face>& getFaces() const;
    std::set<std::shared_ptr<Vertex> > getVertices();
};

class TriangleGroup : public FaceGroup
{
  protected:
    void beforePaint(int shadingType);
    std::shared_ptr<FaceGroup> getFaceGroupPointer();
  public:
    std::shared_ptr<FaceGroup> clone() const;
    void paint(int shadingType) const;
};

class QuadGroup : public FaceGroup
{
  protected:
    void beforePaint(int shadingType);
    std::shared_ptr<FaceGroup> getFaceGroupPointer();
  public:
    std::shared_ptr<FaceGroup> clone() const;
    void paint(int shadingType) const;
};

struct Color
{
  union {
    struct {
      GLfloat r, g, b, a;
    };
    GLfloat v [4];
  };
  Color(GLfloat r_, GLfloat g_, GLfloat b_, GLfloat a_):
    r(r_), g(g_), b(b_), a(a_) {};
  Color(GLfloat r_, GLfloat g_, GLfloat b_):
    Color(r_, g_, b_, 1) {};
};

struct Material
{
  GLfloat shininess;
  Color ambient;
  Color diffuse;
  Color specular;
  Color emission;

  Material():
    ambient(Color(.2, .2, .2)),
    diffuse(Color(.8, .8, .8)),
    specular(Color(0, 0, 0)),
    emission(Color(0, 0, 0)),
    shininess(0) {};

  void apply() const;
};

class Figure : public Transformable, public Paintable, public std::enable_shared_from_this<Figure>
{
  typedef Transformable super;
  protected:
    std::vector<std::shared_ptr<FaceGroup> > groups;
    bool finalized;
  public:
    Material material;
    Vector center;
    GLfloat boundingRadius;
    Figure();
    Figure(const Figure& other);
    void applyTransformations();
    void cloneInto(std::shared_ptr<Figure> target) const;
    void finalize();
    std::set<std::shared_ptr<Vertex> > getVertices();
    Figure& addGroup(std::shared_ptr<FaceGroup> group);
    bool collidesWith(const Figure& other) const;
    bool insideBounds(const Figure& other, float padding) const;
    bool insideBounds(const Figure& other) const;
  protected:
    virtual std::shared_ptr<Figure> getFigurePointer() =0;
    std::vector<Face> getFaces();
    const std::vector<Face> getFaces() const;
    void calculateBoundingSphere();
  private:
    void resetNormals(std::set<std::shared_ptr<Vertex> >& vertices);
    void normalizeNormals(std::set<std::shared_ptr<Vertex> >& vertices);
    void paintMethod(int shadingType);
    void beforePaint(int shadingType);
    void setNormals();
};
#endif
