#include "util.h"
#include "custom_figures.h"
#include "imageloader.h"
#include <memory>

using namespace std;
/*
 * Pyramid
 */

Pyramid::Pyramid()
{
  shared_ptr<Vertex> baseA = make_shared<Vertex>(-1, 0, 1);
  shared_ptr<Vertex> baseB = make_shared<Vertex>(1, 0, 1);
  shared_ptr<Vertex> baseC = make_shared<Vertex>(1, 0, -1);
  shared_ptr<Vertex> baseD = make_shared<Vertex>(-1, 0, -1);
  shared_ptr<Vertex> top = make_shared<Vertex>(0, 1, 0);

  Face base;
  base.addVertex(baseA)
    .addVertex(baseB)
    .addVertex(baseC)
    .addVertex(baseD);

  Face sideA;
  sideA.addVertex(baseA)
    .addVertex(baseB)
    .addVertex(top);

  Face sideB;
  sideB.addVertex(baseB)
    .addVertex(baseC)
    .addVertex(top);

  Face sideC;
  sideC.addVertex(baseC)
    .addVertex(baseD)
    .addVertex(top);

  Face sideD;
  sideD.addVertex(baseD)
    .addVertex(baseA)
    .addVertex(top);

  shared_ptr<FaceGroup> baseGroup = make_shared<QuadGroup>();
  baseGroup->addFace(base);

  shared_ptr<FaceGroup> sidesGroup = make_shared<TriangleGroup>();
  sidesGroup->
    addFace(sideA)->
    addFace(sideB)->
    addFace(sideC)->
    addFace(sideD);

  addGroup(baseGroup);
  addGroup(sidesGroup);
  finalize();
}

shared_ptr<Figure> Pyramid::getFigurePointer()
{
  return shared_from_this();
}

shared_ptr<Pyramid> Pyramid::clone() const
{
  shared_ptr<Pyramid> result = make_shared<Pyramid>();
  cloneInto(result->getFigurePointer());
  return result;
}

shared_ptr<Figure> FishTank::getFigurePointer()
{
  return shared_from_this();
}

bool FishTank::swarmFits(const Swarm& swarm) const
{
  bool result = true;
  for (auto& aFish : swarm.fish)
    if (!aFish.insideBounds(*this, 5)) {
      result = false;
      break;
    }

  return result;
}

bool FishTank::sharkFits(const Shark& shark) const
{
  return shark.insideBounds(*this, 5);
}

FishTank::FishTank(const Swarm& swarm, const Shark& shark)
{
  shared_ptr<Vertex> a1 = make_shared<Vertex>(0, 0, 0);
  shared_ptr<Vertex> b1 = make_shared<Vertex>(3, 0, 0);
  shared_ptr<Vertex> c1 = make_shared<Vertex>(3, 0, 2);
  shared_ptr<Vertex> d1 = make_shared<Vertex>(0, 0, 2);

  shared_ptr<Vertex> a2 = make_shared<Vertex>(0, 2, 0);
  shared_ptr<Vertex> b2 = make_shared<Vertex>(3, 2, 0);
  shared_ptr<Vertex> c2 = make_shared<Vertex>(3, 2, 2);
  shared_ptr<Vertex> d2 = make_shared<Vertex>(0, 2, 2);

  Face top;
  top.addVertex(a2)
    .addVertex(d2)
    .addVertex(c2)
    .addVertex(b2);

  Face bottom;
  bottom.addVertex(a1)
    .addVertex(b1)
    .addVertex(c1)
    .addVertex(d1);

  Face front;
  front.addVertex(d1)
    .addVertex(c1)
    .addVertex(c2)
    .addVertex(d2);

  Face back;
  back.addVertex(a1)
    .addVertex(a2)
    .addVertex(b2)
    .addVertex(b1);

  Face left;
  left.addVertex(a1)
    .addVertex(d1)
    .addVertex(d2)
    .addVertex(a2);

  Face right;
  right.addVertex(b1)
    .addVertex(b2)
    .addVertex(c2)
    .addVertex(c1);

  shared_ptr<FaceGroup> group = make_shared<QuadGroup>();
  group->addFace(top)
    ->addFace(bottom)
    ->addFace(front)
    ->addFace(back)
    ->addFace(left)
    ->addFace(right);

  addGroup(group);
  material.shininess = 50;
  material.ambient = Color(0, 0, 0, 0.4);
  material.diffuse = Color(0, 0, 0, 0.4);
  material.specular = Color(0.5, 0.5, 1);
  material.emission = Color(0, 0, 0.2);
  finalize();
  Vector swarmCenter = swarm.center();
  while (!swarmFits(swarm) || !sharkFits(shark)) {
    scale(Vector(1.1, 1.1, 1.1));
    applyTransformations();
    move(swarmCenter - center);
    applyTransformations();
  }
}

Plant::Plant(const char* filename, int width, int height)
{
  GLuint texture;
  Image *image = loadBMP(filename);

  glGenTextures( 1, &texture );
  glBindTexture( GL_TEXTURE_2D, texture );
  glTexImage2D(GL_TEXTURE_2D,
      0,
      GL_RGB,
      image->width,
      image->height,
      0,
      GL_RGB,
      GL_UNSIGNED_BYTE,
      image->pixels);

  delete image;

  material.diffuse = Color(0,0,0,0);
  material.ambient = Color(0,0,0,0);
  material.specular = Color(0,0,0,0);

  shared_ptr<Vertex> a1 = make_shared<Vertex>(0, 0, 0);
  shared_ptr<Vertex> b1 = make_shared<Vertex>(1, 0, 0);
  shared_ptr<Vertex> c1 = make_shared<Vertex>(1, 0, 0.1);
  shared_ptr<Vertex> d1 = make_shared<Vertex>(0, 0, 0.1);

  shared_ptr<Vertex> a2 = make_shared<Vertex>(0, 3, 0);
  shared_ptr<Vertex> b2 = make_shared<Vertex>(1, 3, 0);
  shared_ptr<Vertex> c2 = make_shared<Vertex>(1, 3, 0.1);
  shared_ptr<Vertex> d2 = make_shared<Vertex>(0, 3, 0.1);

  Face top;
  top.addVertex(a2)
    .addVertex(d2)
    .addVertex(c2)
    .addVertex(b2);

  Face bottom;
  bottom.addVertex(a1)
    .addVertex(b1)
    .addVertex(c1)
    .addVertex(d1);

  Face front;
  front.addVertex(d1)
    .addVertex(c1)
    .addVertex(c2)
    .addVertex(d2);
  top.useTexture = true;
  top.texture = texture;

  Face back;
  back.addVertex(a1)
    .addVertex(a2)
    .addVertex(b2)
    .addVertex(b1);
  top.useTexture = true;
  top.texture = texture;

  Face left;
  left.addVertex(a1)
    .addVertex(d1)
    .addVertex(d2)
    .addVertex(a2);

  Face right;
  right.addVertex(b1)
    .addVertex(b2)
    .addVertex(c2)
    .addVertex(c1);

  shared_ptr<FaceGroup> group = make_shared<QuadGroup>();
  group->addFace(top)
    ->addFace(bottom)
    ->addFace(front)
    ->addFace(back)
    ->addFace(left)
    ->addFace(right);

  addGroup(group);
}

shared_ptr<Figure> Plant::getFigurePointer()
{
  return shared_from_this();
}
