#include "swarm.h"
#include <vector>
#include <cfloat>

using namespace std;

bool Swarm::collides(const Fish& someFish) const
{
  bool result = false;
  for (auto& aFish : fish)
    if (aFish.collidesWith(someFish)) {
      result = true;
      break;
    }
  return result;
}

Swarm::Swarm(int size_) : size(size_)
{
  for (int i = 0; i < size; i++)
    spawnFish();
}

Vector Swarm::center() const
{
  vector<Vector> centers;
  for (auto& aFish : fish)
    centers.push_back(aFish.center);
  if (centers.size() == 0)
    return Vector(0, 0, 0);
  return Vector::average(centers);
}

Fish Swarm::findClosestFish(const Fish& someFish) const
{
  Fish closest;
  float closestDistance = FLT_MAX;
  for (auto& aFish : fish) {
    float distance = someFish.center.distance(aFish.center);
    if (distance != 0 && distance < closestDistance) {
      closest = aFish;
      closestDistance = distance;
    }
  }
  return closest;
}

void Swarm::move(const Figure& bounds, Shark& shark)
{
  float moveAmount;
  float separationWeight = 10;
  float alignmentWeight = 4;
  float cohesionWeight = 1;
  Vector swarmCenter = center();
  vector<Vector> directions;
  vector<Fish>::iterator aFish = fish.begin();
  while (aFish != fish.end() || fish.empty()) {
    moveAmount = 2;
    bool sharkBehind = (Vector(shark.getMouth() + aFish->center) & shark.direction) > 0;
    float distanceToShark = aFish->center.distance(shark.getMouth());
    bool closeToEdge = !(aFish->insideBounds(bounds, 2 * aFish->boundingRadius));
    if (distanceToShark < shark.boundingRadius * 2 && sharkBehind) {
      moveAmount = 6;
      if (closeToEdge)
        aFish->turnTowards(bounds.center, 8);
      else
        aFish->turnAwayFrom(shark.center, 8);
    } else if (closeToEdge) {
      aFish->turnTowards(bounds.center, 6);
    }
    if (!closeToEdge) {
      Fish closest = findClosestFish(*aFish);
      aFish->turnAwayFrom(closest.center, separationWeight / (aFish->center.distance(closest.center) + 0.01));
      aFish->turnInDirection(averageDirection, alignmentWeight);
      aFish->turnTowards(swarmCenter, cohesionWeight);
    }
    aFish->moveForward(moveAmount * (aFish->boundingRadius / 6));
    directions.push_back(aFish->direction);
    if (distanceToShark < aFish->boundingRadius / 2 && sharkBehind) {
      aFish = fish.erase(aFish);
      shark.scale(Vector(1.001, 1.001, 1.001));
    }
    else
      ++aFish;
  }
  if (fish.size() < size / 1.5)
    while (fish.size() < size)
      spawnFish();

  averageDirection = Vector::average(directions);
}

void Swarm::spawnFish()
{
  Fish aFish;
  GLfloat xScale = static_cast <GLfloat> (rand()) / static_cast <GLfloat> (RAND_MAX) + 0.5;
  GLfloat yScale = static_cast <GLfloat> (rand()) / static_cast <GLfloat> (RAND_MAX) + 0.5;
  GLfloat zScale = static_cast <GLfloat> (rand()) / static_cast <GLfloat> (RAND_MAX) + 0.5;
  aFish.scale(Vector(xScale, yScale, zScale));
  aFish.applyTransformations();
  aFish.material.shininess = 128;
  aFish.material.diffuse = Color(xScale - 0.5, yScale - 0.5, 1);
  aFish.material.specular = Color(0, 0, 1);
  aFish.move(center());
  aFish.applyTransformations();

  Vector offset;
  while (collides(aFish)) {
    offset[rand() % 3] += (static_cast <GLfloat> (rand()) / static_cast <GLfloat> (RAND_MAX) - 0.5) * 2;
    aFish.move(offset);
    aFish.applyTransformations();
  }
  fish.push_back(aFish);
}

void Swarm::beforePaint(int shadingType)
{
  
}
void Swarm::paintMethod(int shadingType)
{
  for (auto& aFish : fish)
    aFish.paint(shadingType);
}

