#include "translib.h"
#include <cmath>
#include <exception>
#include <vector>

#if !defined M_PI
#define M_PI 3.14159265358979323846
#endif

/*
 * Vector4
 */

float& Vector4::operator[](int index)
{
  return v[index];
}

const float& Vector4::operator[](const int index) const
{
  return v[index];
}

bool Vector4::operator==(const Vector4& other) const
{
  for (int i = 0; i < 4; i++)
    if ((*this)[i] != other[i])
      return false;
  return true;
}

bool Vector4::operator!=(const Vector4& other) const
{
  return !(*this == other);
}

Vector4& Vector4::operator=(const Vector4& other)
{
  for (int i = 0; i < 4; i++)
    (*this)[i] = other[i];
  return *this;
}

Vector4 Vector4::operator+(const Vector4& other) const
{
  Vector4 result;
  for (int i = 0; i < 4; i++)
    result[i] = (*this)[i] + other[i];
  return result;
}

Vector4 Vector4::operator-(const Vector4& other) const
{
  Vector4 result;
  for (int i = 0; i < 4; i++)
    result[i] = (*this)[i] - other[i];
  return result;
}

Vector4 Vector4::operator*(const float factor) const
{
  Vector4 result;
  for (int i = 0; i < 4; i++)
    result[i] = (*this)[i] * factor;
  return result;
}

Vector4 Vector4::operator/(const float divident) const
{
  Vector4 result;
  for (int i = 0; i < 4; i++)
    result[i] = (*this)[i] / divident;
  return result;
}

float Vector4::operator&(const Vector4& other) const
{
  float sum = 0;
  for (int i = 0; i < 4; i++)
    sum += (*this)[i] * other[i];
  return sum;
}

Vector4 Vector4::operator-() const
{
  return *this * -1;
}

Vector4& Vector4::operator+=(const Vector4& other)
{
  return *this = *this + other;
}

Vector4& Vector4::operator-=(const Vector4& other)
{
  return *this = *this - other;
}

Vector4& Vector4::operator*=(const float factor)
{
  return *this = *this * factor;
}

Vector4& Vector4::operator/=(const float divident)
{
  return *this = *this / divident;
}
/* end Vector4 */

/*
 * Matrix
 */

Vector4& Matrix4::operator[](int index)
{
  switch (index) {
  case 0:
    return a1;
  case 1:
    return a2;
  case 2:
    return a3;
  case 3:
    return a4;
  default:
    throw std::exception();
  }
}

const Vector4& Matrix4::operator[](const int index) const
{
  switch (index) {
  case 0:
    return a1;
  case 1:
    return a2;
  case 2:
    return a3;
  case 3:
    return a4;
  default:
    throw std::exception();
  }
}

bool Matrix4::operator==(const Matrix4& other) const
{
  bool result = true;
  for (int i = 0; i < 4; i++)
    if ((*this)[i] != other[i]) {
      result = false;
      break;
    }
  return result;
}

bool Matrix4::operator!=(const Matrix4& other) const
{
  return !(*this == other);
}

Matrix4& Matrix4::operator=(const Matrix4& other)
{
  for (int i = 0; i < 4; i++)
    (*this)[i] = other[i];
  return *this;
}

Matrix4 Matrix4::operator*(const Matrix4& other) const
{
  Matrix4 result;
  for (int i = 0; i < 4; i++)
    for (int k = 0; k < 4; k++)
      result[i][k] = (*this)[i] & Vector4(other[0][k], other[1][k], other[2][k], other[3][k]);

  return result;
}

Vector4 Matrix4::operator*(const Vector4& v) const
{
  Vector result;
  for (int i = 0; i < 4; i++)
    result[i] = (*this)[i] & v;
  return result;
}

Matrix4 Matrix4::operator*(const float factor) const
{
  Matrix4 result = *this;
  for (int i = 0; i < 4; i++)
    result[i] *= factor;
  return result;
}

Matrix4 Matrix4::operator+(const Matrix4& other) const
{
  Matrix4 result = *this;
  for (int i = 0; i < 4; i++) {
    result[i] += other[i];
  }
  return result;
}

Matrix4 Matrix4::operator-(const Matrix4& other) const
{
  Matrix4 result = *this;
  for (int i = 0; i < 4; i++) {
    result[i] -= other[i];
  }
  return result;
}

Matrix4& Matrix4::operator*=(const Matrix4& other)
{
  return *this = *this * other;
}

Matrix4& Matrix4::operator*=(float factor)
{
  return *this = *this * factor;
}

Matrix4& Matrix4::operator+=(const Matrix4& other)
{
  return *this = *this + other;
}

Matrix4& Matrix4::operator-=(const Matrix4& other)
{
  return *this = *this - other;
}

Matrix4 Matrix4::transpose() const
{
  Matrix4 result;
  for (int i = 0; i < 4; i++)
    for (int k = 0; k < 4; k++)
      result[i][k] = (*this)[k][i];
  return result;
}

float Matrix4::determinant() const
{
  Matrix4 m = *this;
  return
    m[0][3]*m[1][2]*m[2][1]*m[3][0] -
    m[0][2]*m[1][3]*m[2][1]*m[3][0] -
    m[0][3]*m[1][1]*m[2][2]*m[3][0] +
    m[0][1]*m[1][3]*m[2][2]*m[3][0] +
    m[0][2]*m[1][1]*m[2][3]*m[3][0] -
    m[0][1]*m[1][2]*m[2][3]*m[3][0] -
    m[0][3]*m[1][2]*m[2][0]*m[3][1] +
    m[0][2]*m[1][3]*m[2][0]*m[3][1] +
    m[0][3]*m[1][0]*m[2][2]*m[3][1] -
    m[0][0]*m[1][3]*m[2][2]*m[3][1] -
    m[0][2]*m[1][0]*m[2][3]*m[3][1] +
    m[0][0]*m[1][2]*m[2][3]*m[3][1] +
    m[0][3]*m[1][1]*m[2][0]*m[3][2] -
    m[0][1]*m[1][3]*m[2][0]*m[3][2] -
    m[0][3]*m[1][0]*m[2][1]*m[3][2] +
    m[0][0]*m[1][3]*m[2][1]*m[3][2] +
    m[0][1]*m[1][0]*m[2][3]*m[3][2] -
    m[0][0]*m[1][1]*m[2][3]*m[3][2] -
    m[0][2]*m[1][1]*m[2][0]*m[3][3] +
    m[0][1]*m[1][2]*m[2][0]*m[3][3] +
    m[0][2]*m[1][0]*m[2][1]*m[3][3] -
    m[0][0]*m[1][2]*m[2][1]*m[3][3] -
    m[0][1]*m[1][0]*m[2][2]*m[3][3] +
    m[0][0]*m[1][1]*m[2][2]*m[3][3];
}

Matrix4 Matrix4::inverse() const
{
  Matrix4 m = *(this);
  Matrix4 i;
   i[0][0] = m[1][2]*m[2][3]*m[3][1] -
             m[1][3]*m[2][2]*m[3][1] +
             m[1][3]*m[2][1]*m[3][2] -
             m[1][1]*m[2][3]*m[3][2] -
             m[1][2]*m[2][1]*m[3][3] +
             m[1][1]*m[2][2]*m[3][3];

   i[0][1] = m[0][3]*m[2][2]*m[3][1] -
             m[0][2]*m[2][3]*m[3][1] -
             m[0][3]*m[2][1]*m[3][2] +
             m[0][1]*m[2][3]*m[3][2] +
             m[0][2]*m[2][1]*m[3][3] -
             m[0][1]*m[2][2]*m[3][3];

   i[0][2] = m[0][2]*m[1][3]*m[3][1] -
             m[0][3]*m[1][2]*m[3][1] +
             m[0][3]*m[1][1]*m[3][2] -
             m[0][1]*m[1][3]*m[3][2] -
             m[0][2]*m[1][1]*m[3][3] +
             m[0][1]*m[1][2]*m[3][3];

   i[0][3] = m[0][3]*m[1][2]*m[2][1] -
             m[0][2]*m[1][3]*m[2][1] -
             m[0][3]*m[1][1]*m[2][2] +
             m[0][1]*m[1][3]*m[2][2] +
             m[0][2]*m[1][1]*m[2][3] -
             m[0][1]*m[1][2]*m[2][3];

   i[1][0] = m[1][3]*m[2][2]*m[3][0] -
             m[1][2]*m[2][3]*m[3][0] -
             m[1][3]*m[2][0]*m[3][2] +
             m[1][0]*m[2][3]*m[3][2] +
             m[1][2]*m[2][0]*m[3][3] -
             m[1][0]*m[2][2]*m[3][3];

   i[1][1] = m[0][2]*m[2][3]*m[3][0] -
             m[0][3]*m[2][2]*m[3][0] +
             m[0][3]*m[2][0]*m[3][2] -
             m[0][0]*m[2][3]*m[3][2] -
             m[0][2]*m[2][0]*m[3][3] +
             m[0][0]*m[2][2]*m[3][3];

   i[1][2] = m[0][3]*m[1][2]*m[3][0] -
             m[0][2]*m[1][3]*m[3][0] -
             m[0][3]*m[1][0]*m[3][2] +
             m[0][0]*m[1][3]*m[3][2] +
             m[0][2]*m[1][0]*m[3][3] -
             m[0][0]*m[1][2]*m[3][3];

   i[1][3] = m[0][2]*m[1][3]*m[2][0] -
             m[0][3]*m[1][2]*m[2][0] +
             m[0][3]*m[1][0]*m[2][2] -
             m[0][0]*m[1][3]*m[2][2] -
             m[0][2]*m[1][0]*m[2][3] +
             m[0][0]*m[1][2]*m[2][3];

   i[2][0] = m[1][1]*m[2][3]*m[3][0] -
             m[1][3]*m[2][1]*m[3][0] +
             m[1][3]*m[2][0]*m[3][1] -
             m[1][0]*m[2][3]*m[3][1] -
             m[1][1]*m[2][0]*m[3][3] +
             m[1][0]*m[2][1]*m[3][3];

   i[2][1] = m[0][3]*m[2][1]*m[3][0] -
             m[0][1]*m[2][3]*m[3][0] -
             m[0][3]*m[2][0]*m[3][1] +
             m[0][0]*m[2][3]*m[3][1] +
             m[0][1]*m[2][0]*m[3][3] -
             m[0][0]*m[2][1]*m[3][3];

   i[2][2] = m[0][1]*m[1][3]*m[3][0] -
             m[0][3]*m[1][1]*m[3][0] +
             m[0][3]*m[1][0]*m[3][1] -
             m[0][0]*m[1][3]*m[3][1] -
             m[0][1]*m[1][0]*m[3][3] +
             m[0][0]*m[1][1]*m[3][3];

   i[2][3] = m[0][3]*m[1][1]*m[2][0] -
             m[0][1]*m[1][3]*m[2][0] -
             m[0][3]*m[1][0]*m[2][1] +
             m[0][0]*m[1][3]*m[2][1] +
             m[0][1]*m[1][0]*m[2][3] -
             m[0][0]*m[1][1]*m[2][3];

   i[3][0] = m[1][2]*m[2][1]*m[3][0] -
             m[1][1]*m[2][2]*m[3][0] -
             m[1][2]*m[2][0]*m[3][1] +
             m[1][0]*m[2][2]*m[3][1] +
             m[1][1]*m[2][0]*m[3][2] -
             m[1][0]*m[2][1]*m[3][2];

   i[3][1] = m[0][1]*m[2][2]*m[3][0] -
             m[0][2]*m[2][1]*m[3][0] +
             m[0][2]*m[2][0]*m[3][1] -
             m[0][0]*m[2][2]*m[3][1] -
             m[0][1]*m[2][0]*m[3][2] +
             m[0][0]*m[2][1]*m[3][2];

   i[3][2] = m[0][2]*m[1][1]*m[3][0] -
             m[0][1]*m[1][2]*m[3][0] -
             m[0][2]*m[1][0]*m[3][1] +
             m[0][0]*m[1][2]*m[3][1] +
             m[0][1]*m[1][0]*m[3][2] -
             m[0][0]*m[1][1]*m[3][2];

   i[3][3] = m[0][1]*m[1][2]*m[2][0] -
             m[0][2]*m[1][1]*m[2][0] +
             m[0][2]*m[1][0]*m[2][1] -
             m[0][0]*m[1][2]*m[2][1] -
             m[0][1]*m[1][0]*m[2][2] +
             m[0][0]*m[1][1]*m[2][2];

   return i * (1/m.determinant());
}

Matrix4 Matrix4::to3() const
{
  Matrix4 result = *this;
  for (int i = 0; i < 3; i++)
    result[i][3] = 0;
  result[3] = Vector();
  return result;
}
/* end Matrix */


/*
 * Vector
 */
Vector Vector::operator*(const Vector& other) const
{
  return Vector (
    y * other.z - z * other.y,
    z * other.x - x * other.z,
    x * other.y - y * other.x
  );
}

Vector Vector::operator*(float factor) const
{
  return Vector(static_cast<Vector4>(*this) * factor);
}

Vector Vector::operator/(float factor) const
{
  return Vector(static_cast<Vector4>(*this) / factor);
}

Vector& Vector::operator*=(const Vector& other)
{
  return *this = *this * other;
}

Vector& Vector::operator*=(float factor)
{
  return *this = *this * factor;
}

Vector& Vector::operator/=(float factor)
{
  return *this = *this / factor;
}

float Vector::operator&(const Vector& other) const
{
  float sum = 0;
  for (int i = 0; i < 3; i++)
    sum += (*this)[i] * other[i];
  return sum;
}

float Vector::length() const
{
  return sqrt(x*x + y*y + z*z);
}

Vector Vector::normal() const
{
  return *this / length();
}

Vector& Vector::normalize()
{
  return *this /= length();
}

float Vector::distance(const Vector& other) const
{
  return Vector(*this - other).length();
}

Vector Vector::average(const std::vector<Vector>& vectors)
{
  Vector sum;
  for (auto& vec : vectors)
    sum += vec;
  return Vector(sum / vectors.size());
}
/* end Vector */


/*
 * Transformable
 */
void Transformable::updateNormalTransformations()
{
  normalTransformations = transformations.to3().inverse().transpose();
}

Matrix4 Transformable::translationMatrix(const Vector& offset) const
{
  float x = offset.x, y = offset.y, z = offset.z;
  return Matrix4(
    1, 0, 0, x,
    0, 1, 0, y,
    0, 0, 1, z,
    0, 0, 0, 1
  );
}

Matrix4 Transformable::rotationMatrix(const Vector& axis, float angle, const Vector& center) const
{
  float u = axis.x, v = axis.y, w = axis.z;
  float uu = u*u, uv = u*v, uw = u*w;
  float vv = v*v, vw = v*w;
  float ww = w*w;
  angle = angle * (M_PI / 180);
  float cosA = cos(angle);
  float sinA = sin(angle);

  float a = center.x, b = center.y, c = center.z;
  float au = a*u, av = a*v, aw = a*w;
  float bu = b*u, bv = b*v, bw = b*w;
  float cu = c*u, cv = c*v, cw = c*w;

  return Matrix4(
    uu+(vv+ww)*cosA    , uv*(1-cosA)-w*sinA , uw*(1-cosA)+v*sinA , (a*(vv+ww)-u*(bv+cw))*(1-cosA)+(bw-cv)*sinA ,
    uv*(1-cosA)+w*sinA , vv+(uu+ww)*cosA    , vw*(1-cosA)-u*sinA , (b*(uu+ww)-v*(au+cw))*(1-cosA)+(cu-aw)*sinA ,
    uw*(1-cosA)-v*sinA , vw*(1-cosA)+u*sinA , ww+(uu+vv)*cosA    , (c*(uu+vv)-w*(au+bv))*(1-cosA)+(av-bu)*sinA ,
    0                  , 0                  , 0                  , 1
  );
}

Matrix4 Transformable::scalingMatrix(const Vector& factors) const
{
  float x = factors.x, y = factors.y, z = factors.z;
  return Matrix4(
    x, 0, 0, 0,
    0, y, 0, 0,
    0, 0, z, 0,
    0, 0, 0, 1
  );
}

Transformable::Transformable()
{
  transformations = Matrix4(1);
}

void Transformable::move(const Vector& offset)
{
  transformations = translationMatrix(offset) * transformations;
  if (transformations[3] != Vector())
    transformations[2] = transformations[2];
  updateNormalTransformations();
}

void Transformable::scale(const Vector& factors)
{
  transformations = scalingMatrix(factors) * transformations;
  updateNormalTransformations();
}

void Transformable::rotate(const Vector& axis, float angle, const Vector& center)
{
  transformations = rotationMatrix(axis.normal(), angle, center) * transformations;
  updateNormalTransformations();
}

void Transformable::rotate(const Vector& axis, float angle)
{
  rotate(axis, angle, Vector());
}

void Transformable::clearTransformations()
{
  transformations = Matrix4(1);
  normalTransformations = Matrix4(1);
}
